from . import helpers


def _transform(command, *args):
    """Apply command's transformation function (if any) to given arguments.

    Arguments:
    command -- the command description dict
    *args -- command arguments
    """
    if "value_transform" in command:
        return command["value_transform"](*args)
    return args if len(args) > 1 else args[0]


def choice_handler(command, choice):
    """Returns command bytes from choices commands.

    Arguments:
    command -- the command description dict
    choice -- the choosen value
    """
    if choice not in command["choices"]:
        raise ValueError(
                "value must be one of [%s]" %
                helpers.choices_to_string(command["choices"]))
    value = command["choices"][choice]
    value = _transform(command, value)
    return helpers.merge_bytes(command["command"], value)


def rgbcolor_handler(command, *args):
    """Returns command bytes from RGB color commands.

    Arguments:
    command -- the command description dict
    *args -- the color as a string (color name or hexadecimal RGB), or as R, G
             and B int (e.g. fn(cmd, "red"), fn(cmd, "#ff0000"),
             fn(cmd, 255, 0, 0))
    """
    color = (0x00, 0x00, 0x00)
    if len(args) == 3:
        for value in args:
            if type(value) != int or value < 0 or value > 255:
                raise ValueError("Not a valid color %s" % str(args))
        color = args
    elif len(args) == 1 and type(args[0]) == str and helpers.is_color(args[0]):
        color = helpers.color_string_to_rgb(args[0])
    else:
        raise ValueError("Not a valid color %s" % str(args))
    color = _transform(command, *color)
    return helpers.merge_bytes(command["command"], color)


def rgbcolorshift_handler(command, colors, speed=200):
    """Returns command bytes from RGB color commands.

    Arguments:
    command -- the command description dict
    colors -- the colors as an array of string (color name or hexadecimal RGB),
              or as an array of array of R, G and B int (e.g. fn(cmd, ["red"]),
              fn(cmd, ["#ff0000"]), fn(cmd, [[255, 0, 0]]))

    Keyword arguments:
    speed -- the color shift speed (default 200ms)
    """
    parsed_colors = []
    for color in colors:
        if type(color) in (list, tuple) and len(color) == 3:
            parsed_colors.extend(color)
        elif type(color) == str and helpers.is_color(color):
            parsed_colors.extend(helpers.color_string_to_rgb(color))
        else:
            raise ValueError("Not a valid color %s" % str(color))
    parsed_colors, speed = _transform(command, parsed_colors, speed)
    speed = helpers.uint_to_little_endian_bytearray(speed, 2)
    return helpers.merge_bytes(command["command"], parsed_colors, speed)


def rgbuniversal_handler(command, colors, positions, speed, triggers):
    """ Returns command bytes from RGB color commands.
    It crafts packets on a per mouse basis, by reading the format given
    in "rgbuniversal_format" in the command dict. If a single color is given,
    it creates a steady color command. If multiple colors are given, it creates
    a colorshift command. If triggermask is nonzero, a reactive command is
    created.

    Arguments:
    command -- the command description dict
    colors -- the colors as an array of string (color name or hexadecimal RGB)
    positions -- cumulative color positions as an array of hex strings
    speed -- the colorshift cycle time as a string (milliseconds)
    triggers -- trigger button mask as a hex string
    """
    colors = list(map(helpers.color_string_to_rgb, colors))
    positions = map(lambda x: int(x, 16), positions)
    speed = 5000 if speed.lower() == "x" else int(speed, 10)
    triggers = 0 if triggers.lower() == "x" else int(triggers, 16)

    rgb_format = command["rgbuniversal_format"]
    header = [0] * rgb_format["header_len"]

    for led_id in rgb_format["led_id"]:
        header[led_id] = command["led_id"]

    speed_pos, speed_len = rgb_format["speed"], rgb_format["speed_len"]
    speed = helpers.uint_to_little_endian_bytearray(speed, speed_len)

    header[speed_pos:speed_pos + speed_len] = speed
    header[rgb_format["repeat"]] = 1 if len(colors) == 1 or triggers > 0 else 0
    header[rgb_format["triggers"]] = triggers
    header[rgb_format["point_count"]] = len(colors) + 1

    # data segment format:
    # color1:color1:pos1:...:colorn:posn:color1:pos1
    data = colors[0]

    for color, pos in zip(colors, positions):
        data = helpers.merge_bytes(data, color, pos)

    remaining = 0xff - sum(positions)
    if remaining < 0:
        raise ValueError("The cumulative position offsets must be < 255 (is %d)" % sum(positions)) # noqa

    # inserts first color at the end for smooth colorshifting
    data = helpers.merge_bytes(data, colors[0], remaining)

    return helpers.merge_bytes(command["command"], header, data)


def range_handler(command, value):
    """Returns command bytes from range commands.

    Arguments:
    command -- the command description dict
    value -- the choosen value
    """
    if not command["range_min"] <= value <= command["range_max"]:
        raise ValueError("Value %i not in range [%i, %i]" % (
            value,
            command["range_min"],
            command["range_max"]
            ))
    if ("range_increment" in command
            and value % command["range_increment"] != 0):
        raise ValueError("Value %i is not a multiple of %i" % (
            value,
            command["range_increment"]
            ))
    value = _transform(command, value)
    return helpers.merge_bytes(command["command"], value)


def none_handler(command):
    """Returns command bytes for command with not arguments.

    Arguments:
    command -- the command description dict
    """
    return command["command"]


def hotsbtnmap_handler(command, value):
    """Returns command bytes for HotS button map.

    Arguments:
    command -- the command description dict
    value -- the choosen value
    """
    olist = helpers.hotsbtnmap_to_list(value)
    if len(olist) != 8:
        raise ValueError("Please provide 8 keys to be mapped.")
    value = _transform(command, value)
    return helpers.merge_bytes(command["command"], *olist)



from .. import usbhid


rival600 = {
    "name": "SteelSeries Rival 600 (Experimental)",

    "vendor_id": 0x1038,
    "product_id": 0x1724,
    "interface_number": 0,

    "rgbuniversal_format": {
        "header_len": 28,   # Number of bytes in header excluding command bytes
        "led_id": [0, 5],   # Index(es) of LED ID (unsure why rival600 have 2)
        "speed": 6,         # Index of the colorshift speed field
        "speed_len": 2,     # How many bytes the speed field takes up
        "repeat": 22,       # Index of the repeat flag
        "triggers": 23,     # Index of the trigger button mask field
        "point_count": 27,  # Index of the color count field
    },

    "commands": {

        "set_sensitivity1": {
            "description": "Set sensitivity preset 1",
            "cli": ["-s", "--sensitivity1"],
            "command": [0x03, 0x00, 0x01],
            "suffix": [0x00, 0x42],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int((x / 100) - 1),
            "default": 800,
        },

        "set_sensitivity2": {
            "description": "Set sensitivity preset 2",
            "cli": ["-S", "--sensitivity2"],
            "command": [0x03, 0x00, 0x02],
            "suffix": [0x00, 0x42],
            "value_type": "range",
            "range_min": 100,
            "range_max": 12000,
            "range_increment": 100,
            "value_transform": lambda x: int((x / 100) - 1),
            "default": 1600,
        },

        "set_polling_rate": {
            "description": "Set polling rate in Hz",
            "cli": ["-p", "--polling-rate"],
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "set_wheel_color": {
            "description": "Set the wheel backlight color(s) and effects",
            "cli": ["-C", "--wheel-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x0,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_logo_color": {
            "description": "Set the logo backlight color(s) and effects",
            "cli": ["-c", "--logo-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x1,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_left_strip_top_color": {
            "description": "Set the color(s) and effects of the left LED strip upper section", # noqa
            "cli": ["-0", "--lstrip-top-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x2,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_left_strip_mid_color": {
            "description": "Set the color(s) and effects of the left LED strip middle section", # noqa
            "cli": ["-1", "--lstrip-mid-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x4,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_left_strip_bottom_color": {
            "description": "Set the color(s) and effects of the left LED strip bottom section", # noqa
            "cli": ["-2", "--lstrip-bottom-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x6,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_right_strip_top_color": {
            "description": "Set the color(s) and effects of the right LED strip upper section", # noqa
            "cli": ["-3", "--rstrip-top-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x3,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_right_strip_mid_color": {
            "description": "Set the color(s) and effects of the right LED strip mid section", # noqa
            "cli": ["-4", "--rstrip-mid-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x5,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "set_right_strip_bottom_color": {
            "description": "Set the color(s) and effects of the right LED strip bottom section", # noqa
            "cli": ["-5", "--rstrip-bottom-color"],
            "command": [0x05, 0x00],
            "report_type": usbhid.HID_REPORT_TYPE_FEATURE,
            "value_type": "rgbuniversal",
            "led_id": 0x7,
            "default": (["red", "green", "blue"], ["0", "54", "54"], "x", "x")
        },

        "save": {
            "description": "Save the configuration to the mouse memory",
            "cli": None,
            "command": [0x09, 0x00],
            "value_type": None,
        },
    }
}



from .rival import rival

rival300acerpredator = {
    "name": "Acer Predator Gaming Mouse (Rival 300)",

    "vendor_id": 0x1038,
    "product_id": 0x1714,
    "interface_number": 0,

    "commands": rival["commands"]
}



from __future__ import print_function
import sys
import platform

from . import cli
from . import debug
from .version import VERSION
from . import list_supported_mice, list_available_mice, get_mouse


def get_first_available_mouse():
    available_mice = list(list_available_mice())
    if not available_mice:
        return None
    mouse = available_mice[0]
    try:
        return get_mouse(
                mouse.vendor_id,
                mouse.product_id
                )
    except IOError as error:
        print("W: The following mouse was found but rivalcfg was not able to open it:")  # noqa
        print("  * mouse: %s (%4X:%4X)" % mouse)
        print("  * error: %s" % error)
        print("\nPlease check that no other application is controlling this mouse and try to:")  # noqa
        print("  * unplug the mouse from the USB port,")
        print("  * wait few seconds,")
        print("  * and plug the mouse to the USB port again.\n")
        return None


def _print_debug_info():
    debug.log("Rivalcfg %s" % VERSION)
    debug.log("Python version: %s" % platform.python_version())
    debug.log("OS: %s" % platform.system())
    if debug.DRY:
        debug.log("Dry run enabled")
    if debug.get_debug_profile():
        debug.log("Forced profile: %04X:%04X" % debug.get_debug_profile())
        debug.log("Targeted device: %04X:%04X" % debug.get_debug_device())


def _print_compatible_mice():
    """Prints mice currently supported by this software."""
    print("\n".join(["%-50s %04X:%04X   %s" % (
            mouse.name,
            mouse.vendor_id,
            mouse.product_id,
            "(plugged)" if mouse in list_available_mice() else ""
        ) for mouse in list_supported_mice()]))


def main(argv=sys.argv[1:]):
    _print_debug_info()

    mouse = get_first_available_mouse()
    debug.log("Selected mouse: %s" % mouse)

    profile = mouse.profile if mouse else None
    parser = cli.generate_cli(profile)
    options, args = parser.parse_args(argv)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    # Standard options

    if options.list:
        _print_compatible_mice()
        sys.exit(0)

    # Mouse specific options

    if not mouse:
        print("Error: no compatible mouse found.")
        sys.exit(1)

    if hasattr(options, "reset") and options.reset:
        mouse.set_default()

    for option, value in list(options.__dict__.items()):
        if option in ["list", "reset"] or value is None:
            continue
        if (profile["commands"][option]["value_type"] in ["choice", "range"]
                and value.isdigit()):
            value = int(value)
        if not type(value) in (list, tuple):
            value = [value]
        count = getattr(mouse, option)(*value)
        if count < 0:
            print("An error occured when updating mouse settings!")
            sys.exit(1)

    if hasattr(mouse, "save"):
        count = mouse.save()
        if count < 0:
            print("An error occured when saving mouse settings!")
            sys.exit(1)


if __name__ == "__main__":
    main(sys.argv[1:])



"""
This module contains functions to get information about mice and to get an
object to manipulate a specific mouse.

Usage and examples
------------------

Changing the color of the first mouse found::

    import rivalcfg

    mouse = rivalcfg.get_first_mouse()
    mouse.set_color("red")

Listing all available mice and changing their color::

    import rivalcfg

    mice_info = rivalcfg.list_available_mice()

    for mouse_info in mice_info:
        mouse = rivalcfg.get_mouse(mouse_info.vendor_id, mouse_info.product_id)
        mouse.set_color(red)

.. NOTE::

    In previous examples, we assumed that only Rival 100 mice were connected to
    the computer: the methods available on the ``mouse`` object depends on the
    mouse model.


Public API
----------
"""


import collections

from . import usbhid
from .version import VERSION
from .profiles import mice_profiles
from .mouse import Mouse


#: Tuple containing information about a specific mouse
MouseInfo = collections.namedtuple(
        "MouseInfo", ("name", "vendor_id", "product_id"))


def list_supported_mice():
    """Returns the list of all mice supported by this software.

    :rtype: list of :py:class:`rivalcfg.MouseInfo`
    """
    return (MouseInfo(
            name=profile["name"],
            vendor_id=profile["vendor_id"],
            product_id=profile["product_id"],
            ) for profile in mice_profiles)


def list_available_mice():
    """Return the list of supported mice currently availale (plugged to the
    computer).

    :rtype: list of :py:class:`rivalcfg.MouseInfo`
    """
    for mouse in list_supported_mice():
        if usbhid.is_device_plugged(mouse.vendor_id, mouse.product_id):
            yield mouse


def get_mouse_profile(vendor_id, product_id):
    """Returns the profile (informations and list of available commands) of the
    given mouse.

    :param int vendor_id: The vendor id (currenlty this should be set to
                          0x1038)
    :param int product_id: The product id of the mouse (e.g. 0x1710 for the
                           Rival 300)

    :rtype: dict
    """
    for profile in mice_profiles:
        if (profile["vendor_id"] == vendor_id
                and profile["product_id"] == product_id):
            return profile
    return None


def get_mouse(vendor_id, product_id):
    """Returns a class that allows you to manipulaite the given mouse if it is
    available, else returns None.

    :param int vendor_id: The vendor id (currenlty this should be set to
                          0x1038)
    :param int product_id: The product id of the mouse (e.g. 0x1710 for the
                           Rival 300)

    :rtype: :py:class:`rivalcfg.mouse.Mouse`
    """
    profile = get_mouse_profile(vendor_id, product_id)
    if not profile:
        return None
    return Mouse(profile)


def get_first_mouse():
    """Returns the first available mouse if any mouse is connected, else
    returns None.

    :rtype: :py:class:`rivalcfg.mouse.Mouse`
    """
    for mouse in list_available_mice():
        return get_mouse(mouse.vendor_id, mouse.product_id)
    return None


__all__ = [
        "VERSION",
        "list_supported_mice",
        "list_available_mice",
        "get_mouse_profile",
        "get_mouse",
        "get_first_mouse",
        ]



kanav2 = {
    "name": "SteelSeries Kana V2",

    "vendor_id": 0x1038,
    "product_id": 0x137a,
    "interface_number": 0,

    "commands": {

        "set_sensitivity1": {
            "description": "Set sensitivity preset 1",
            "cli": ["-s", "--sensitivity1"],
            "command": [0x03, 0x01],
            "value_type": "choice",
            "choices": {
                400: 0x08,
                800: 0x07,
                1200: 0x06,
                1600: 0x05,
                2000: 0x04,
                2400: 0x03,
                3200: 0x02,
                4000: 0x01,
            },
            "default": 800,
        },

        "set_sensitivity2": {
            "description": "Set sensitivity preset 2",
            "cli": ["-S", "--sensitivity2"],
            "command": [0x03, 0x02],
            "value_type": "choice",
            "choices": {
                400: 0x08,
                800: 0x07,
                1200: 0x06,
                1600: 0x05,
                2000: 0x04,
                2400: 0x03,
                3200: 0x02,
                4000: 0x01,
            },
            "default": 1600,
        },

        "set_polling_rate": {
            "description": "Set polling rate in Hz",
            "cli": ["-p", "--polling-rate"],
            "command": [0x04, 0x00],
            "value_type": "choice",
            "choices": {
                125: 0x04,
                250: 0x03,
                500: 0x02,
                1000: 0x01,
            },
            "default": 1000,
        },

        "set_led_intensity1": {
            "description": "Set LED intensity preset 1",
            "cli": ["-i", "--intensity1"],
            "command": [0x05, 0x01],
            "value_type": "choice",
            "choices": {
                "off": 0x01,
                "low": 0x02,
                "medium": 0x03,
                "high": 0x04,
            },
            "default": "off",
        },

        "set_led_intensity2": {
            "description": "Set LED intensity preset 2",
            "cli": ["-I", "--intensity2"],
            "command": [0x05, 0x02],
            "value_type": "choice",
            "choices": {
                "off": 0x01,
                "low": 0x02,
                "medium": 0x03,
                "high": 0x04,
            },
            "default": "high",
        },

        "save": {
            "description": "Save the configuration to the mouse memory",
            "cli": None,
            "command": [0x09, 0x00],
            "value_type": None,
        },

    },

}



from . import usbhid
from . import debug
from . import helpers
from . import command_handlers


REPORT_TYPE_TO_HIDAPI_FUNCTION = {
    0x02: "write",
    0x03: "send_feature_report"
    }


class Mouse:

    """Generic class to handle any supported mouse. The methods availabled on
    this class depends of the loaded profile.

    :param dict profile: the mouse profile to load (``rivalcfg.profiles.*``)
    """

    profile = None
    _device = None

    def __init__(self, profile):
        """Contructor."""
        self.profile = profile
        self._device = usbhid.open_device(
                profile["vendor_id"],
                profile["product_id"],
                profile["interface_number"])

    def set_default(self):
        """Sets all option to their factory values."""
        for command in self.profile["commands"]:
            if "default" in self.profile["commands"][command]:
                default = self.profile["commands"][command]["default"]
                if not type(default) in [list, tuple]:
                    default = [default]
                getattr(self, command)(*default)

    def _device_write(self, bytes_, report_type=usbhid.HID_REPORT_TYPE_OUTPUT):
        """Writes bytes to the device.

        :param bytes bytes_: bytes to write
        :param int report_type: the HID Repport Type (0x02: output (default),
                                0x03: feature)

        :returns int: the number of bytes written, or -1 on failure.
        """
        report_id = 0x00
        if debug.DEBUG:
            debug.log_bytes_hex(
                    "Mouse._device_write [wValue]",
                    [report_type, report_id])
            debug.log_bytes_hex("Mouse._device_write   [data]", bytes_)
        bytes_ = helpers.merge_bytes(report_id, bytes_)
        report_function = getattr(
                self._device,
                REPORT_TYPE_TO_HIDAPI_FUNCTION[report_type])
        return report_function(bytearray(bytes_))

    def __getattr__(self, name):
        if name not in self.profile["commands"]:
            raise AttributeError("There is no command named '%s'" % name)

        command = self.profile["commands"][name]
        handler = "%s_handler" % str(command["value_type"]).lower()

        report_type = usbhid.HID_REPORT_TYPE_OUTPUT
        if "report_type" in command:
            report_type = command["report_type"]

        suffix = []
        if "suffix" in command:
            suffix = command["suffix"]

        if command["value_type"] == "rgbuniversal":
            if "rgbuniversal_format" in self.profile:
                debug.log("rgbuniversal_format found")
                command["rgbuniversal_format"] = \
                    self.profile["rgbuniversal_format"]
            else:
                raise Exception("rgbuniversal value type was specified, but the format is not defined for %s" % self.profile["name"]) # noqa

        if not hasattr(command_handlers, handler):
            raise Exception("There is not handler for the '%s' value type" % command["value_type"])  # noqa

        def _exec_command(*args):
            bytes_ = getattr(command_handlers, handler)(command, *args)
            bytes_ = helpers.merge_bytes(bytes_, suffix)
            return self._device_write(bytes_, report_type)

        return _exec_command

    def __repr__(self):
        return "<Mouse %s (%04X:%04X:%02X)>" % (
                self.profile["name"],
                self.profile["vendor_id"],
                self.profile["product_id"],
                self.profile["interface_number"])

    def __str__(self):
        return self.__repr__()

    def __del__(self):
        if self._device:
            self._device.close()



from .rival100 import rival100

rival100dota2edition = {
    "name": "SteelSeries Rival 100 Dota 2 Edition",

    "vendor_id": 0x1038,
    "product_id": 0x170c,
    "interface_number": 0,

    "commands": rival100["commands"]

}



from .rival import rival
from .rival100 import rival100
from .rival100dota2edition import rival100dota2edition
from .rival110 import rival110
from .rival300 import rival300
from .rival300acerpredator import rival300acerpredator
from .rival300csgofadeedition import rival300csgofadeedition
from .rival300csgohyperbeastedition import rival300csgohyperbeastedition
from .rival300dota2edition import rival300dota2edition
from .rival300hpomen import rival300hpomen
from .rival310 import rival310
from .rival500 import rival500
from .rival600 import rival600
from .hotssenseiraw import hotssenseiraw
from .kanav2 import kanav2

mice_profiles = [
    rival,
    rival100,
    rival100dota2edition,
    rival110,
    rival300,
    rival300acerpredator,
    rival300csgofadeedition,
    rival300csgohyperbeastedition,
    rival300dota2edition,
    rival300hpomen,
    rival310,
    rival500,
    rival600,
    hotssenseiraw,
    kanav2,
]



import re


NAMED_COLORS = {
    "white":   (0xFF, 0xFF, 0xFF),
    "silver":  (0xC0, 0xC0, 0xC0),
    "gray":    (0x80, 0x80, 0x80),
    "black":   (0x00, 0x00, 0x00),
    "red":     (0xFF, 0x00, 0x00),
    "maroon":  (0x80, 0x00, 0x00),
    "yellow":  (0xFF, 0xFF, 0x00),
    "olive":   (0x80, 0x80, 0x00),
    "lime":    (0x00, 0xFF, 0x00),
    "green":   (0x00, 0x80, 0x00),
    "aqua":    (0x00, 0xFF, 0xFF),
    "teal":    (0x00, 0x80, 0x80),
    "blue":    (0x00, 0x00, 0xFF),
    "navy":    (0x00, 0x00, 0x80),
    "fuchsia": (0xFF, 0x00, 0xFF),
    "purple":  (0x80, 0x00, 0x80),

    # Rival 300 CS:GO Fade Edition presets
    "preset1": (0xFF, 0x52, 0x00),
    "preset2": (0x1D, 0xC5, 0xFF),
    "preset3": (0x64, 0x03, 0xFC),
    "preset4": (0xFF, 0xF2, 0x00),
    "preset5": (0xFF, 0x00, 0x00),
}

# Button support for Heroes of the Storm (Sensei Raw)
NAMED_KEYS = {
    "lctrl":   [0x10, 0xE0, 0x00],
    "lshift":  [0x10, 0xE1, 0x00],
    "lalt":    [0x10, 0xE2, 0x00],
    "lcmd":    [0x10, 0xE3, 0x00],
    "rctrl":   [0x10, 0xE4, 0x00],
    "rshift":  [0x10, 0xE5, 0x00],
    "ralt":    [0x10, 0xE6, 0x00],
    "rcmd":    [0x10, 0xE7, 0x00],
    "pgdn":    [0x10, 0x4B, 0x00],
    "pgup":    [0x10, 0x4E, 0x00],
    # home and end in b/w pageup and down i.e. 4C, 4D ;
    # don't remember the order.
    "senst":   [0x30, 0x00, 0x00],
    "mouse1":  [0x01, 0x00, 0x00],
    "mouse2":  [0x02, 0x00, 0x00],
    "mouse3":  [0x03, 0x00, 0x00],
    "mouse4":  [0x04, 0x00, 0x00],
    "mouse5":  [0x05, 0x00, 0x00],
    "mouse6":  [0x06, 0x00, 0x00],
    "mouse7":  [0x07, 0x00, 0x00],
    "mouse8":  [0x08, 0x00, 0x00],
}


def is_color(string):
    """Checks if the given string is a valid color.

    Arguments:
    string -- the string to check
    """
    if string in NAMED_COLORS:
        return True
    if re.match(r"^#?[0-9a-f]{3}([0-9a-f]{3})?$", string, re.IGNORECASE):
        return True
    return False


def is_hex(string):
    try:
        int(string, 16)
        return True
    except ValueError:
        return False


def color_string_to_rgb(color_string):
    """Converts the color string into an RGB tuple.

    Arguments:
    color_string -- the string to converts

    Returns:
    an (R, G, B) tuple
    """
    # Named color
    if color_string in NAMED_COLORS:
        return NAMED_COLORS[color_string]
    # #f00 or #ff0000 -> f00 or ff0000
    if color_string.startswith("#"):
        color_string = color_string[1:]
    # f00 -> ff0000
    if len(color_string) == 3:
        color_string = color_string[0] * 2 + color_string[1] * 2 + color_string[2] * 2  # noqa
    # ff0000 -> (255, 0, 0)
    return (
        int(color_string[0:2], 16),
        int(color_string[2:4], 16),
        int(color_string[4:], 16)
        )


def hotsbtnmap_to_list(kstring):
    """Converts a sequence of keymaps to a list of keymap settings for each button.
    Returns a list of strings

    Arguments:
    kstring -- The key sequence
    """
    kstring = kstring.split()
    outlist = []

    # the list of commands to be outputted
    if not len(kstring) == 8:
        raise ValueError("Invalid length of argument to Set Button Commands!")
    for x in range(0, 8):
        dd = kstring[x].lower()
        if dd in NAMED_KEYS:
            outlist.append(NAMED_KEYS[kstring[x]])
            continue

        if len(dd) == 1 and 0 <= ord(dd) - ord("a") <= 25:
            outlist.append([0x10, ord(dd) - ord("a") + 4, 0x00])
            continue

        if len(dd) == 1 and 0 <= ord(dd) - ord("0") <= 9:
            if(dd == "0"):
                outlist.append([0x10, 0x27, 0x00])
            else:
                outlist.append([0x10, ord(dd) - ord("1") + 0x1E, 0x00])
            continue

        raise ValueError("Invalid entry key name")

    return outlist  # It is a list of lists.


def choices_to_list(choices):
    """Transforms choices dict to an ordered string list.

    Arguments:
    choices -- the dict containing available choices
    """
    return list(map(str, sorted(choices.keys(), key=lambda v: v if type(v) == int else -1)))  # noqa


def choices_to_string(choices):
    """Transforms choices dict to a printable string.

    Arguments:
    choices -- the dict containing available choices
    """
    return ", ".join(choices_to_list(choices))


def merge_bytes(*args):
    """Returns a single list of bytes from given bytes and list of bytes.

    Arguments:
    *args -- bytes and list of bytes to merge
    """
    result = []
    for arg in args:
        if type(arg) in [list, tuple]:
            result.extend(arg)
        else:
            result.append(arg)
    return result


def uint_to_little_endian_bytearray(number, size):
    """Converts an unsigned interger to a little endian bytearray.

    Arguments:
    number -- the number to convert
    size -- the length of the target bytearray
    """
    if number > (2 ** (8 * size) - 1):
        raise ValueError("Integer overflow")
    nle = [0] * size
    for i in range(size):
        nle[i] = number >> i*8 & 0xFF
    return nle
